/*
	ShowPixelValues plugin for Avisynth -- lets you view the 
	Y U and V values (or R G and B) within each pixel

	Copyright (C) 2002-2003 Simon Walters, All Rights Reserved

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Simon Walters
	siwalters@hotmail.com

V0.3 - simple make pixels bigger and label
V1.0 - 1st release YUY2 only
V1.1 - Add in RGB32
V1.2 - Expose Bright, Dim and Threshold settings for displaying levels.
V1.3 - underline Y1 value in YUY2 mode
V1.4 - Add ShowPixels parameter to give option of having a clear display 
		(Just YUY2 and no longer showpixel values of borders)
V1.4a - Fixed Problems crashes with non-standard image sizes 
V1.4b - Cosmetic changes to certain numbers in font and fix bug with buf variable
V1.4c - Made ShowPixels=False work in RGB and YUY2 
*/
#include <stdio.h>
#include "windows.h"
#include "avisynth.h"
#include "info.h"

#include "memory.h"

class ShowPixelValues : public GenericVideoFilter {
	bool ShowValues, ShowArea, ShowPixels;
	int Left, Top,  Bright, Dim, Threshold;
	int Width, Height;
	int CellSize;
//	PClip resized;

	

public:
	ShowPixelValues(PClip _child,
	int _Left, int _Top, bool _ShowValues, bool _ShowArea, int _Bright, int _Dim, int _Threshold,
		bool _ShowPixels, IScriptEnvironment* env) :
	GenericVideoFilter(_child),
		 Left(_Left),Top(_Top),ShowValues(_ShowValues),
		  ShowArea(_ShowArea), Bright(_Bright), Dim(_Dim), Threshold(_Threshold), ShowPixels(_ShowPixels)

	{
	// constructor code
		if (!(vi.IsRGB32()||vi.IsYUY2()))
			env->ThrowError("ShowPixelValues: input to filter must be in YUY2 or RGB32 colorspace");
		if (!(((Left>>1)<<1) == Left))
			env->ThrowError("ShowPixelValues: value of Left parameter must be even");


		CellSize = 16;
		Height = vi.height / CellSize;
		Top = min(Top,(vi.height- Height));
		Width = ((vi.width / CellSize)>>1)<<1;
		Left = (Left>>1)<<1;
		Left = min(Left,(vi.width-Width));

		if (ShowPixels == false) {
			Threshold=256;
		}

//   AVSValue args[3] = { child, (vi.width*2), (vi.height) }; 

//	resized = env->Invoke("BilinearResize",AVSValue(args,3)).AsClip(); 
	vi.width = vi.width*2;


	}

// Destructor code
	ShowPixelValues::~ShowPixelValues(void) {
// delete[] resized;
//	delete resizedPVF;
	}

    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};



PVideoFrame __stdcall ShowPixelValues::GetFrame(int n, IScriptEnvironment* env) {


		PVideoFrame src = child->GetFrame(n, env);
		PVideoFrame dst = env->NewVideoFrame(vi);

//		PVideoFrame test = resized->GetFrame(n,env);
//		env->MakeWritable(&test);

		
		const unsigned char* srcp = src->GetReadPtr();
		unsigned char* dstp = dst->GetWritePtr();
//		unsigned char* testp = test->GetWritePtr();		
		
		const int dst_pitch = dst->GetPitch();
		const int dst_width = dst->GetRowSize();
		const int dst_height = dst->GetHeight();

		const int src_pitch = src->GetPitch();
		const int src_width = src->GetRowSize();
		const int src_height = src->GetHeight();
		
//		const int test_pitch = test->GetPitch();
//		const int test_width = test->GetRowSize();
//		const int test_height = test->GetHeight();
		
		int dst_modulo;
		dst_modulo = dst_pitch - dst_width;
	
		
		int w, h;
//		return dst;

		// YUY2 code
		if (vi.IsYUY2())
		{
			srcp = src->GetReadPtr();
			dstp = dst->GetWritePtr();
			
			// WRITE src to right half dst
			for (h=0;h < (src_height);h++) {
				for (w=0;w < (src_width);w+=1) {
					*(dstp + w + (src_width) + (h * dst_pitch)) = *(srcp + w + (h * src_pitch));
				}
			}
			// empty left half dst
			for (h=0;h < (src_height);h++) {
				for (w=0;w < (src_width);w+=4) {
					*((Pixel32 *)(dstp + w + (h * dst_pitch))) = 0x80A780A7;
				}
			}

			if (ShowArea)
			{
				for (h=Top;h < Top +2 ;h++) {
					for (w=(((Left>>1)<<2));w < (((Left>>1)<<2) - (0) +((src_width>>6)<<2));w+=4) {
						*((Pixel32 *)(dstp + w + src_width + (h * dst_pitch))) = 0xDD6ACA6A;
					}
				}

				for (h=Top - 2 + Height;h < Top - 0 + Height;h++) {
					for (w=(((Left>>1)<<2));w < (((Left>>1)<<2) - (0) +((src_width>>6)<<2));w+=4) {
						*((Pixel32 *)(dstp + w + src_width + (h * dst_pitch))) = 0xDD6ACA6A;
					}
				}
	
				for (w=(((Left>>1)<<2)+ 0);w < (((Left>>1)<<2) + (4));w+=4) {
					for (h=Top+2;h < Top - 2 + Height;h++) {
						*((Pixel32 *)(dstp + w + src_width + (h * dst_pitch))) = 0xDD6ACA6A;
					}
				}
				for (w=(((Left>>1)<<2) - (4) +((src_width>>6)<<2));w < (((Left>>1)<<2) - (0) +((src_width>>6)<<2));w+=4) {
					for (h=Top+2;h < Top - 2 + Height;h++) {
						*((Pixel32 *)(dstp + w + src_width + (h * dst_pitch))) = 0xDD6ACA6A;
					}
				}
			}
			// end ShowArea

			// write examined area from right dst to left dst
			for (h=Top;h < Top + Height;h++) {
				int dummy = 0;
				for (w=((Left>>1)<<2);w < ((Left>>1)<<2) +((src_width>>6)<<2);w+=1) {
					*(dstp + w - ((Left>>1)<<2) + ((h-Top) * dst_pitch)) = *(dstp + w +src_width + (h * dst_pitch));
				dummy++;
				}
				int dummy2 = dummy;
			}


			// Separate Y1 and Y2 values into their own 32bit "pixels"
			// i.e double 16 130 18 131 becomes 16 130 16 131        18 130 18 131
			for (h=0;h < (src_height);h++) {
				for (w = ((src_width>>6)<<2)-4;w>=0;w-=4) {
					
					*(dstp + (w*2) + 7 + (h * dst_pitch)) = *(dstp +w + 3 + (h * dst_pitch));
					*(dstp + (w*2) + 6 + (h * dst_pitch)) = *(dstp  +(w+2) + (h * dst_pitch));
					*(dstp + (w*2) + 5 + (h * dst_pitch)) = *(dstp  +w + 1 + (h * dst_pitch));
					*(dstp + (w*2) + 4 + (h * dst_pitch)) = *(dstp  +(w+2) + (h * dst_pitch));
					*(dstp + (w*2) + 3 + (h * dst_pitch)) = *(dstp  +w + 3 + (h * dst_pitch));
					*(dstp + (w*2) + 2 + (h * dst_pitch)) = *(dstp  +w + (h * dst_pitch));
					*(dstp + (w*2) + 1 + (h * dst_pitch)) = *(dstp  +w + 1 + (h * dst_pitch));
					*(dstp + (w*2) +     (h * dst_pitch)) = *(dstp  + w +     (h * dst_pitch));
					
					
				}
			}

			// multiply horizontally
			for (h=0;h < (src_height);h++) {
				for (w=(((src_width>>2)<<2)-4);w >=0;w-=4) {
					int ww =((w>>5)<<2);
					*(dstp + w + (h * dst_pitch)) = *(dstp + ww + (h * dst_pitch));
					*(dstp + w + 1 + (h * dst_pitch)) = *(dstp + ww + 1 +(h * dst_pitch));
					*(dstp + w + 2 + (h * dst_pitch)) = *(dstp + ww + 2 + (h * dst_pitch));
					*(dstp + w + 3 + (h * dst_pitch)) = *(dstp + ww + 3 + (h * dst_pitch));
				}
			}

			// multiply vertically				
			for (h=(dst_height - 1);h >= 0;h-=1) {
				for (w=0;w < ((src_width>>2)<<2);w+=1) {
					*(dstp + w + (h * dst_pitch)) = *(dstp + w + ((h>>4) * dst_pitch));
				}
			}

		
//return dst;	
			// write out values to frame
			if (ShowValues == true) 
			{
				for (h=0;h < (Height * CellSize);h+=CellSize)
				{
//					for (w=(CellSize * 2 * 4);w < ((Width - 2) * CellSize * 4);w+=64)

					for (w=0;w < (Width*32);w+=32)
					{
						int YValue= *(dstp + w + (h * dst_pitch));
						unsigned int CValue= *(dstp + w + 1 + (int)((w % 64) / 16) + (h * dst_pitch));
						*(dstp + w + (h * dst_pitch)) = 235;	
						
						// Blank out pixels if needed
						if (ShowPixels == false) 
						{
							int x,y;
							
							for (y=0;y<16;y++)
							{
								for (x=0;x<32;x+=2) 
								{
									*(dstp + w + ((h+y) * dst_pitch)+x) =0;
									*(dstp + w + ((h+y) * dst_pitch)+x+1) =128;
								}
							}
							
						} // End Blank out pixels


						char buf[4];
						sprintf(buf, "%d",YValue);
						int ylevel = Bright;
						if (YValue > Threshold)
							ylevel = Dim;
						
						DrawString(dst, (w), (h+1), buf,ylevel, vi);
						if ((w % 64) == 0)
						{
							sprintf(buf, "___",YValue);
							DrawString(dst, (w), (h+2), buf,ylevel, vi);
						}
						
						//				char buf[5];
						sprintf(buf, "%d",CValue);
						
						DrawString(dst, (w), (h+9), buf, ylevel,vi);
						
						



					} 
				}
			}

			// Draw dividing line
			for (h=0;h < (src_height);h++) {
				for (w=(src_width-4);w < (src_width);w+=4) {
					*((Pixel32 *)(dstp + w + (h * dst_pitch))) = 0x80A780A7;
				}
			}


		}
		else if (vi.IsRGB32())
		{
			srcp = src->GetReadPtr();
			dstp = dst->GetWritePtr();
			
			// WRITE src to right half dst
			for (h=0;h < (src_height);h++) {
				for (w=0;w < (src_width);w+=1) {
					*(dstp + w + (src_width) + (h * dst_pitch)) = *(srcp + w + (h * src_pitch));
				}
			}

			
			// empty left half dst
			for (h=0;h < (src_height);h++) {
				for (w=0;w < (src_width);w+=4) {
					*((Pixel32 *)(dstp + w + (h * dst_pitch))) = 0xB0B0B0;
				}
			}

			// Invert image (cause RGB is "upside down")
			for (h=0;h < (dst_height/2);h++) {
				for (w=0;w < (dst_width);w+=1) {
					unsigned char temp = *(dstp + w + (h * dst_pitch));
					*(dstp + w + (h * dst_pitch))	= *(dstp + w + ((dst_height-h-1) * dst_pitch));
					*(dstp + w + ((dst_height-h-1) * dst_pitch)) = temp;
				}
			}

			if (ShowArea)
			{
				for (h=Top;h < Top +2 ;h++) {
					for (w=(((Left)<<2));w < (((Left)<<2) - (0) +(Width<<2));w+=4) {
						*((Pixel32 *)(dstp + w + src_width + (h * dst_pitch))) = 0xFF00FF;
					}
				}

				for (h=Top - 2 + Height;h < Top - 0 + Height;h++) {
					for (w=(((Left)<<2));w < (((Left)<<2) - (0) +(Width<<2));w+=4) {
						*((Pixel32 *)(dstp + w + src_width + (h * dst_pitch))) = 0xFF00FF;
					}
				}
	
				for (w=(((Left)<<2)+ 0);w < (((Left)<<2) + (8));w+=4) {
					for (h=Top+2;h < Top - 2 + Height;h++) {
						*((Pixel32 *)(dstp + w + src_width + (h * dst_pitch))) = 0xFF00FF;
					}
				}
				for (w=(((Left)<<2) - (8) +(Width<<2));w < (((Left)<<2) - (0) +(Width<<2));w+=4) {
					for (h=Top+2;h < Top - 2 + Height;h++) {
						*((Pixel32 *)(dstp + w + src_width + (h * dst_pitch))) = 0xFF00FF;
					}
				}
			}
			// end ShowArea


			// write examined area from right dst to left dst
			for (h=Top;h < Top + Height;h++) {
				int dummy = 0;
				for (w=((Left)<<2);w < ((Left)<<2) + (Width << 2);w+=1) {
					*(dstp + w - ((Left)<<2) + ((h-Top) * dst_pitch)) = *(dstp + w +src_width + (h * dst_pitch));
				dummy++;
				}
				int dummy2 = dummy;
			}

			//multiply horizontally by 16			
			for (h=(dst_height - 1);h >= 0;h-=1) {
				for (w=(((src_width>>2)<<2)-4);w >=0;w-=4) {
					*((Pixel32 *)(dstp + w + (h * dst_pitch))) = *((Pixel32 *)(dstp + ((w>>6)<<2) + (h * dst_pitch)));
				}
			}


			// multiply vertically				
			for (h=(dst_height - 1);h >= 0;h-=1) {
				for (w=0;w < ((src_width>>2)<<2);w+=1) {
					*(dstp + w + (h * dst_pitch)) = *(dstp + w + ((h>>4) * dst_pitch));
				}
			}





			if (ShowValues == true)
			{
				for (h=0;h < (Height * CellSize);h+=CellSize) {
					for (w=0;w < (Width * CellSize * 4);w+=64)
					{
						int BValue= *(dstp + w + (h * dst_pitch));
						int GValue= *(dstp + w + 1 + (h * dst_pitch));
						int RValue= *(dstp + w + 2 + (h * dst_pitch));

						*(dstp + w + (h * dst_pitch)) = 255;
						*(dstp + w + 1 + (h * dst_pitch)) = 255;
						*(dstp + w + 2 + (h * dst_pitch)) = 255;

						// Blank out pixels if needed
						if (ShowPixels == false) 
						{
							int x,y;
							
							for (y=0;y < CellSize;y++)
							{
								for (x=0;x< CellSize;x+=1) 
								{
									*((Pixel32 *)(dstp + w + ((h+y) * dst_pitch)+ (x*4))) =0x000000;
								}
							}
							
						} // End Blank out pixels

						
						char buf[4];
						int rtrn= sprintf(buf, "%d",RValue);
						
						int ylevel = Bright;
	
						if (RValue > Threshold)
							ylevel = Dim;
						
						DrawString(dst, (w), (h+1), buf,(ylevel << 16),vi);
						
						
						sprintf(buf, "%d",GValue);
						
						ylevel = Bright;
						if (GValue > Threshold)
							ylevel = Dim;
						
						DrawString(dst, (w+0), (h+6), buf,(ylevel << 8),vi);
						
						sprintf(buf, "%d",BValue);
						
						ylevel = Bright;
						if (BValue > Threshold)
							ylevel = Dim;
						
						
						DrawString(dst, (w+0), (h+11), buf,ylevel,vi);
					} 
				}
			}

			// Draw dividing line
			for (h=0;h < (src_height);h++) {
				for (w=src_width-8;w < (src_width);w+=4) {
					*((Pixel32 *)(dstp + w + (h * dst_pitch))) = 0xB0B0B0;
				}
			}


			//Flip back again
			for (h=0;h < (dst_height/2);h++) {
				for (w=0;w < (dst_width);w+=1) {
					unsigned char temp = *(dstp + w + (h * dst_pitch));
					*(dstp + w + (h * dst_pitch))	= *(dstp + w + ((dst_height-h-1) * dst_pitch));
					*(dstp + w + ((dst_height-h-1) * dst_pitch)) = temp;
				}
			}
			//end flip back again
			return dst;


		}
	

//		dstp = dst->GetWritePtr();
//		testp = test->GetWritePtr();
/*
		// WRITE Dst to Test
		for (h=0;h < (dst_height);h++) {
			for (w=0;w < (dst_width);w+=1) {
//				*(dstp + w + (h * dst_pitch)) = *(testp + w + (h * test_pitch));
				*(testp + w + (h * test_pitch)) = *(dstp + w + (h * dst_pitch));
			}
		}
*/

		return dst;
//		child = resized;
//		vi = resized->GetVideoInfo();
//		 vi.width = vi.width*2;
//		return test;

	
	
}


AVSValue __cdecl Create_ShowPixelValues(AVSValue args, void* user_data, IScriptEnvironment* env) {
    return new ShowPixelValues(args[0].AsClip(),
		 args[1].AsInt(0),	// Left
		 args[2].AsInt(0),	// Top
		 args[3].AsBool(true),	// ShowValues
		 args[4].AsBool(false),	// ShowArea
		 args[5].AsInt(235),	// Bright
		 args[6].AsInt(160),	// Dim
		 args[7].AsInt(200),	// Threshold
		 args[8].AsBool(true),	// ShowPixels
		 env);
}


extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("ShowPixelValues", "c[Left]i[Top]i[ShowValues]b[ShowArea]b[Bright]i[Dim]i[Threshold]i[ShowPixels]b", Create_ShowPixelValues, 0);
    return "`ShowPixelValues' ShowPixelValues plugin";
}

